from asyncio import create_task, as_completed
from typing import List, Coroutine, Type

from pydantic import BaseModel

from gravybox.exceptions import DataUnavailable


def merge_dicts_and_trim_nones(first: dict, second: dict):
    trimmed_first = {}
    for key, value in first.items():
        if isinstance(value, str):
            if len(value) > 0:
                trimmed_first[key] = value
        elif value is not None:
            trimmed_first[key] = value
    trimmed_second = {}
    for key, value in second.items():
        if isinstance(value, str):
            if len(value) > 0:
                trimmed_second[key] = value
        elif value is not None:
            trimmed_second[key] = value
    result = trimmed_first | trimmed_second
    return result


def all_fields_populated(instance: BaseModel):
    for key, value in instance.model_dump().items():
        if value is None:
            return False
    return True


def no_fields_populated(instance: BaseModel):
    for key, value in instance.model_dump().items():
        if value is not None:
            return False
    return True


class UpstreamCentrifuge:
    def __init__(self, upstream_calls: List[Coroutine], result_model: Type[BaseModel]):
        self.tasks = [create_task(upstream_call) for upstream_call in upstream_calls]
        self.result_model = result_model

    async def activate(self):
        final_result = self.result_model()
        for upstream_call_wrapper in as_completed(self.tasks):
            upstream_result = await upstream_call_wrapper
            if upstream_result is not None:
                final_result_dict = merge_dicts_and_trim_nones(final_result.model_dump(), upstream_result.model_dump())
                final_result = self.result_model.model_validate(final_result_dict)
                if all_fields_populated(final_result):
                    break
        for task in self.tasks:
            if not task.done():
                task.cancel()
        if no_fields_populated(final_result):
            raise DataUnavailable()
        return final_result


class UpstreamSequencer:
    def __init__(self, upstream_calls: List[Coroutine], result_model: Type[BaseModel]):
        self.upstream_calls = upstream_calls
        self.result_model = result_model

    async def activate(self):
        final_result = self.result_model()
        for upstream_call_wrapper in self.upstream_calls:
            upstream_result = await upstream_call_wrapper
            if upstream_result is not None:
                final_result_dict = merge_dicts_and_trim_nones(final_result.model_dump(), upstream_result.model_dump())
                final_result = self.result_model.model_validate(final_result_dict)
                if all_fields_populated(final_result):
                    break
        if no_fields_populated(final_result):
            raise DataUnavailable()
        return final_result


class UpstreamShortCircuit:
    def __init__(self, upstream_calls: List[Coroutine], result_model: Type[BaseModel]):
        self.upstream_calls = upstream_calls
        self.result_model = result_model

    async def activate(self):
        for upstream_call_wrapper in self.upstream_calls:
            upstream_result = await upstream_call_wrapper
            if upstream_result is not None:
                return upstream_result
        raise DataUnavailable()



import httpx
import pytest
from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel
from starlette.testclient import TestClient

from gravybox.exceptions import DataUnavailable
from gravybox.framework import create_app, ThresholdEndpoint
from gravybox.protocol import LinkRequest, LinkResponse, Condition

app = create_app()
threshold_key = "supersecret"
valid_headers = {"x-api-key": threshold_key}

app.add_middleware(ThresholdEndpoint, threshold_key=threshold_key)


class TestModel(BaseModel):
    value_one_plus_one: int
    opposite_of_value_two: bool


class TestRequest(LinkRequest):
    value_one: int
    value_two: bool


class TestResponse(LinkResponse):
    content: TestModel | None


@app.post("/success")
async def success_endpoint(link_request: TestRequest) -> TestResponse:
    result = TestModel(
        value_one_plus_one=link_request.value_one + 1,
        opposite_of_value_two=not link_request.value_two
    )
    return TestResponse(success=True, content=result)


@app.post("/failure")
async def failing_endpoint(link_request: TestRequest) -> TestResponse:
    raise RuntimeError("failing endpoint failed as expected")


@app.post("/data_unavailable")
async def data_unavailable_endpoint(link_request: TestRequest) -> TestResponse:
    raise DataUnavailable()


@app.post("/timeout")
async def timeout_endpoint(link_request: TestRequest) -> TestResponse:
    raise httpx.ReadTimeout("timeout endpoint timed out")


@app.post("/http_exception")
async def http_exception_endpoint(link_request: TestRequest) -> TestResponse:
    raise HTTPException(status_code=404, detail="not found")


@pytest.mark.asyncio
async def test_no_api_key():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="success", value_one=3, value_two=False)
        response = client.post("/success", json=jsonable_encoder(test_request))
        assert response.status_code == 403
        response_model = TestResponse.model_validate(response.json())
        assert response_model.success is False
        assert response_model.error == Condition.authentication_failure.value


@pytest.mark.asyncio
async def test_invalid_api_key():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="success", value_one=3, value_two=False)
        response = client.post("/success", json=jsonable_encoder(test_request), headers={"x-api-key": "nonsense"})
        assert response.status_code == 403
        response_model = TestResponse.model_validate(response.json())
        assert response_model.success is False
        assert response_model.error == Condition.authentication_failure.value


@pytest.mark.asyncio
async def test_framework_no_payload():
    with TestClient(app) as client:
        response = client.post("/success", headers=valid_headers)
        assert response.status_code == 400
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == Condition.invalid_request.value
        assert test_response.content is None


@pytest.mark.asyncio
async def test_framework_no_trace_id_on_link_request():
    with TestClient(app) as client:
        test_request = {
            "value_one": 3,
            "value_two": False
        }
        response = client.post("/success", json=test_request, headers=valid_headers)
        assert response.status_code == 422
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == Condition.invalid_request.value
        assert test_response.content is None


@pytest.mark.asyncio
async def test_framework_malformed_payload():
    with TestClient(app) as client:
        test_request = {
            "trace_id": "malformed_payload",
            "value_one": 3
        }
        response = client.post("/success", json=test_request, headers=valid_headers)
        assert response.status_code == 422
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == Condition.invalid_request.value
        assert test_response.content is None


@pytest.mark.asyncio
async def test_framework_success():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="success", value_one=3, value_two=False)
        response = client.post("/success", json=jsonable_encoder(test_request), headers=valid_headers)
        assert response.status_code == 200
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is True
        assert test_response.error == ""
        assert test_response.content.value_one_plus_one == 4
        assert test_response.content.opposite_of_value_two is True


@pytest.mark.asyncio
async def test_framework_data_unavailable():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="data unavailable", value_one=3, value_two=False)
        response = client.post("/data_unavailable", json=jsonable_encoder(test_request), headers=valid_headers)
        assert response.status_code == 200
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == Condition.data_unavailable.value
        assert test_response.content is None


@pytest.mark.asyncio
async def test_framework_timeout():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="timeout", value_one=3, value_two=False)
        response = client.post("/timeout", json=jsonable_encoder(test_request), headers=valid_headers)
        assert response.status_code == 500
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == Condition.upstream_timeout.value
        assert test_response.content is None


@pytest.mark.asyncio
async def test_framework_failure():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="failure", value_one=3, value_two=False)
        response = client.post("/failure", json=jsonable_encoder(test_request), headers=valid_headers)
        assert response.status_code == 500
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == Condition.unhandled_exception.value
        assert test_response.content is None


@pytest.mark.asyncio
async def test_framework_http_error():
    with TestClient(app) as client:
        test_request = TestRequest(trace_id="failure", value_one=3, value_two=False)
        response = client.post("/http_exception", json=jsonable_encoder(test_request), headers=valid_headers)
        assert response.status_code == 404
        test_response = TestResponse.model_validate(response.json())
        assert test_response.success is False
        assert test_response.error == "not found"
        assert test_response.content is None



from asyncio import sleep

import pytest

from gravybox.betterstack import collect_logger
from gravybox.exceptions import DataUnavailable
from gravybox.protocol import LinkRequest
from gravybox.upstream_call_paradigm import UpstreamCentrifuge, UpstreamSequencer, UpstreamShortCircuit
from test.testkit import sleeping_coroutine, TestTaskResult, none_result_coroutine, failing_coroutine

logger = collect_logger()


@pytest.mark.asyncio
async def test_upstream_centrifuge_single_task():
    link_request = LinkRequest(trace_id="centrifuge")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_two_task():
    link_request = LinkRequest(trace_id="double_centrifuge")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(999, "sleepy", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_failing_task():
    link_request = LinkRequest(trace_id="failing_centrifuge")
    tasks = [
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_none_result_task():
    link_request = LinkRequest(trace_id="none_centrifuge")
    tasks = [
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request),
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_total_failure():
    link_request = LinkRequest(trace_id="total_failure")
    tasks = [
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    with pytest.raises(DataUnavailable):
        await upstream_caller.activate()


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_two_results():
    link_request = LinkRequest(trace_id="centrifuge_merge")
    tasks = [
        sleeping_coroutine(1, None, 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_overwrite():
    link_request = LinkRequest(trace_id="centrifuge_merge_overwrite")
    tasks = [
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="late_precedence", field_two=15, field_three=False)


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_empty_strings():
    link_request = LinkRequest(trace_id="centrifuge_merge_empty_strings")
    tasks = [
        sleeping_coroutine(1, "", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_chaos():
    link_request = LinkRequest(trace_id="centrifuge_chaos")
    tasks = [
        sleeping_coroutine(99, None, None, field_three=True, link_request=link_request),
        failing_coroutine(33, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        none_result_coroutine(3, None, None, field_three=False, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request),
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
        sleeping_coroutine(19, "something_else", 25, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamCentrifuge(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="late_precedence", field_two=15, field_three=False)


@pytest.mark.asyncio
async def test_upstream_sequencer_single_task():
    link_request = LinkRequest(trace_id="sequencer")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_two_task():
    link_request = LinkRequest(trace_id="double_sequencer")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(999, "sleepy", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_sequencing():
    link_request = LinkRequest(trace_id="sequencer_sequencing")
    logger.info("adding first upstream call to tasks")
    tasks = [
        sleeping_coroutine(1, None, 23, field_three=True, link_request=link_request)
    ]
    await sleep(1)
    logger.info("adding second upstream call to tasks")
    tasks.append(
        sleeping_coroutine(2, "test", 23, field_three=True, link_request=link_request)
    )
    await sleep(1)
    logger.info("creating sequencer")
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    await sleep(1)
    logger.info("calling upstreams")
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_failing_task():
    link_request = LinkRequest(trace_id="failing_sequencer")
    tasks = [
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_none_result_task():
    link_request = LinkRequest(trace_id="none_sequencer")
    tasks = [
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_total_failure():
    link_request = LinkRequest(trace_id="total_failure")
    tasks = [
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    with pytest.raises(DataUnavailable):
        await upstream_caller.activate()


@pytest.mark.asyncio
async def test_upstream_sequencer_merge_two_results():
    link_request = LinkRequest(trace_id="sequencer_merge")
    tasks = [
        sleeping_coroutine(1, None, 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_merge_overwrite():
    link_request = LinkRequest(trace_id="sequencer_merge_overwrite")
    tasks = [
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="late_precedence", field_two=15, field_three=False)


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_empty_strings():
    link_request = LinkRequest(trace_id="sequencer_merge_empty_strings")
    tasks = [
        sleeping_coroutine(1, "", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_sequencer_chaos():
    link_request = LinkRequest(trace_id="sequencer_chaos")
    tasks = [
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request),
        none_result_coroutine(3, None, None, field_three=False, link_request=link_request),
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request),
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        sleeping_coroutine(99, None, None, field_three=True, link_request=link_request),
        failing_coroutine(33, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(19, "something_else", 25, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamSequencer(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="late_precedence", field_two=15, field_three=False)


@pytest.mark.asyncio
async def test_upstream_short_circuit_single_task():
    link_request = LinkRequest(trace_id="short_circuit")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_two_task():
    link_request = LinkRequest(trace_id="double_short_circuit")
    tasks = [
        sleeping_coroutine(1, "test", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(999, "sleepy", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_sequencing():
    link_request = LinkRequest(trace_id="short_circuit_sequencing")
    logger.info("adding first upstream call to tasks")
    tasks = [
        sleeping_coroutine(1, None, 23, field_three=True, link_request=link_request)
    ]
    await sleep(1)
    logger.info("adding second upstream call to tasks")
    tasks.append(
        sleeping_coroutine(2, "test", 23, field_three=True, link_request=link_request)
    )
    await sleep(1)
    logger.info("creating short_circuit")
    usptream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    await sleep(1)
    logger.info("calling upstreams")
    result = await usptream_caller.activate()
    assert result == TestTaskResult(field_one=None, field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_failing_task():
    link_request = LinkRequest(trace_id="failing_short_circuit")
    tasks = [
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_none_result_task():
    link_request = LinkRequest(trace_id="none_short_circuit")
    tasks = [
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "test", 23, field_three=True, link_request=link_request)
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="test", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_total_failure():
    link_request = LinkRequest(trace_id="total_failure")
    tasks = [
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    with pytest.raises(DataUnavailable):
        await upstream_caller.activate()


@pytest.mark.asyncio
async def test_upstream_short_circuit_breaker():
    link_request = LinkRequest(trace_id="short_circuit_merge")
    tasks = [
        sleeping_coroutine(1, None, 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one=None, field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_merge_overwrite():
    link_request = LinkRequest(trace_id="short_circuit_merge_overwrite")
    tasks = [
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one=None, field_two=None, field_three=True)


@pytest.mark.asyncio
async def test_upstream_centrifuge_merge_empty_strings():
    link_request = LinkRequest(trace_id="short_circuit_merge_empty_strings")
    tasks = [
        sleeping_coroutine(1, "", 23, field_three=True, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one="", field_two=23, field_three=True)


@pytest.mark.asyncio
async def test_upstream_short_circuit_chaos():
    link_request = LinkRequest(trace_id="short_circuit_chaos")
    tasks = [
        sleeping_coroutine(1, None, None, field_three=True, link_request=link_request),
        failing_coroutine(1, "failure", 333, field_three=False, link_request=link_request),
        none_result_coroutine(3, None, None, field_three=False, link_request=link_request),
        none_result_coroutine(2, None, None, field_three=False, link_request=link_request),
        failing_coroutine(2, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(5, "late_precedence", 15, field_three=False, link_request=link_request),
        sleeping_coroutine(3, "test", None, field_three=True, link_request=link_request),
        sleeping_coroutine(99, None, None, field_three=True, link_request=link_request),
        failing_coroutine(33, "failure", 333, field_three=False, link_request=link_request),
        sleeping_coroutine(19, "something_else", 25, field_three=False, link_request=link_request)
    ]
    upstream_caller = UpstreamShortCircuit(tasks, TestTaskResult)
    result = await upstream_caller.activate()
    assert result == TestTaskResult(field_one=None, field_two=None, field_three=True)



import json
import time
import traceback
import uuid
from contextlib import asynccontextmanager

from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from httpx import ReadTimeout
from starlette.exceptions import HTTPException
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse

from gravybox.betterstack import collect_logger
from gravybox.exceptions import GravyboxException, DataUnavailable
from gravybox.protocol import Condition, GravyboxResponse
from gravybox.requests import AsyncRequestManager

logger = collect_logger()


@asynccontextmanager
async def lifespan(fastapi_app: FastAPI):
    logger.info("( ) starting FastAPI server")
    AsyncRequestManager.initialize()
    try:
        yield
    finally:
        await AsyncRequestManager.shutdown()
    logger.info("(*) shutting down FastAPI server")


class ThresholdEndpoint(BaseHTTPMiddleware):
    def __init__(self, app, threshold_key=None):
        self.threshold_key = threshold_key
        super().__init__(app)

    async def dispatch(self, request, call_next):
        request_key = request.headers.get("x-api-key")
        if request_key == self.threshold_key:
            return await call_next(request)
        else:
            logger.warning("(!) authentication failed", extra={"client_host": request.client.host})
            return JSONResponse(
                status_code=403,
                content=GravyboxResponse(
                    success=False,
                    error=Condition.authentication_failure.value
                ).model_dump()
            )


class LoggingEndpoint(BaseHTTPMiddleware):
    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request, call_next):
        request.state.log_extras = {
            "endpoint": request.scope["path"]
        }
        try:
            payload = await request.json()
            request.state.log_extras["request_json"] = json.dumps(payload)
            request.state.trace_id = payload.get("trace_id", str(uuid.uuid4()))
            request.state.log_extras["trace_id"] = request.state.trace_id
        except Exception as error:
            request.state.log_extras["error_str"] = str(error)
            request.state.log_extras["condition"] = Condition.invalid_request.value
            logger.error("(!) failed to parse request", extra=request.state.log_extras)
            return JSONResponse(
                status_code=400,
                content=GravyboxResponse(
                    success=False,
                    error=request.state.log_extras["condition"]
                ).model_dump()
            )
        logger.info("( ) endpoint receiving request", extra=request.state.log_extras)
        start_time = time.time()
        try:
            response = await call_next(request)
            if "condition" not in request.state.log_extras:
                request.state.log_extras["condition"] = Condition.success.value
        except Exception as error:
            if isinstance(error, GravyboxException):
                request.state.log_extras |= error.log_extras
            request.state.log_extras["error_str"] = str(error)
            request.state.log_extras["traceback"] = traceback.format_exc()
            request.state.log_extras["condition"] = Condition.unhandled_exception.value
            logger.error("(!) endpoint failed with unhandled exception", extra=request.state.log_extras)
            response = JSONResponse(
                status_code=500,
                content=GravyboxResponse(success=False, error=request.state.log_extras["condition"]).model_dump()
            )
        finally:
            request.state.log_extras["status_code"] = response.status_code
            request.state.log_extras["elapsed_time"] = time.time() - start_time
            logger.info("(*) endpoint emitting response", extra=request.state.log_extras)
            return response


def create_app():
    app = FastAPI(lifespan=lifespan)
    app.add_middleware(LoggingEndpoint)

    @app.exception_handler(ReadTimeout)
    async def upstream_timeout_handler(request: Request, error: ReadTimeout):
        request.state.log_extras["condition"] = Condition.upstream_timeout.value
        return JSONResponse(
            status_code=500,
            content=GravyboxResponse(success=False, error=request.state.log_extras["condition"]).model_dump()
        )

    @app.exception_handler(DataUnavailable)
    async def data_unavailable_handler(request: Request, error: DataUnavailable):
        request.state.log_extras["condition"] = Condition.data_unavailable.value
        return JSONResponse(
            status_code=200,
            content=GravyboxResponse(success=False, error=request.state.log_extras["condition"]).model_dump()
        )

    @app.exception_handler(RequestValidationError)
    async def request_validation_handler(request: Request, error: RequestValidationError):
        request.state.log_extras["condition"] = Condition.invalid_request.value
        return JSONResponse(
            status_code=422,
            content=GravyboxResponse(success=False, error=request.state.log_extras["condition"]).model_dump()
        )

    @app.exception_handler(HTTPException)
    async def http_exception_handler(request: Request, error: HTTPException):
        request.state.log_extras["condition"] = error.detail
        return JSONResponse(
            status_code=error.status_code,
            content=GravyboxResponse(success=False, error=request.state.log_extras["condition"]).model_dump()
        )

    return app



from gravybox.betterstack import collect_logger

logger = collect_logger()


class GravyboxException(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.log_extras = {}


class BadStatusCode(GravyboxException):
    def __init__(self, response):
        message = "bad status code"
        super().__init__(message)
        self.response = response
        self.log_extras["status_code"] = response.status_code
        self.log_extras["response_content"] = response.text


class DataUnavailable(GravyboxException):
    def __init__(self):
        message = "data is not available"
        super().__init__(message)


class UnexpectedCondition(GravyboxException):
    def __init__(self, condition):
        message = "encountered unexpected condition"
        super().__init__(message)
        self.log_extras["condition"] = condition
